﻿using System;
using System.Collections.Generic;
using System.Reflection;

namespace ExtensibleObjectProperties
{
	/// <summary>
	/// This class template is alist of attributes (of type ATTRIB)
	/// It is automatically contructed for an object (of type OBJ_TYPE)
	/// Using all class's properties defined with the System.Attribute (of type PROPERTY_SYS_ATTRIB)
	/// </summary>
	/// <typeparam name="ATTRIB">The attribute object which will be contained in the list</typeparam>
	/// <typeparam name="OBJ_TYPE">The object type (class)</typeparam>
	/// <typeparam name="PROPERTY_SYS_ATTRIB">The systtem attribute. 
	/// All object's properties having this System.Attribute will be automatically included in the list.</typeparam>
	public class xBuiltinPropertiesListBase<PROPERTY, OBJ_TYPE, PROPERTY_SYS_ATTRIB> : List<PROPERTY> where PROPERTY : PropertyInfo
	{
		protected string		m_list_name	= "dynamic object attribute list";


		public xBuiltinPropertiesListBase()
			: base()
		{
			init_instance();
		}

		public xBuiltinPropertiesListBase(string list_name)
			: base()
		{
			m_list_name	= list_name;
			init_instance();
		}


		protected void init_instance()
		{
			LoadClassAttributeProperties();
		}


		public bool SetValue(OBJ_TYPE obj, PropertyInfo property_info, object value)
		{
			if( obj == null || property_info == null)
				return false;

			return SetValue( obj, property_info.Name, value);
		}

		/// <summary>
		/// Set property value
		/// </summary>
		/// <param name="obj">The object whoes property value is to be set.</param>
		/// <param name="property_name">The property name</param>
		/// <param name="value">The value to be assigned</param>
		/// <returns>true if successful</returns>
		public bool SetValue(OBJ_TYPE obj, string property_name, object value)
		{
			if( obj == null)
				return false;
			
			PROPERTY		my_property	= this[property_name];

			if( my_property == null)
				return false;

			// is this a readonly property?
			if( my_property.CanWrite == false)
				return false;

			// get the property's data type
			Type	property_type	= my_property.PropertyType;
			object	new_value;

			try
			{
				// try to convert the given object value to property's data type
				new_value	= Convert.ChangeType( value, property_type);
			}
			catch (Exception)
			{
				return false;
			}

			try
			{
				// try to assign the converted value to the object
				my_property.SetValue( obj, new_value, null);
			}
			catch (Exception)
			{
				return false;
			}

			return true;
		}


		/// <summary>
		/// Automatically construct the attribute list
		/// </summary>
		/// <returns>true if successful.</returns>
		public bool LoadClassAttributeProperties()
		{
			Clear();

			Type			obj_type	= typeof(OBJ_TYPE);

			// loop through all class's properties.
			// insert those properties defined with the specified System.Attribute.
			foreach (PropertyInfo p in obj_type.GetProperties())
			{
				// Does this property define our System.Attribute?. If so, add it to the list.
				bool	has_sys_attribute		= Attribute.IsDefined( p, typeof( PROPERTY_SYS_ATTRIB), false);

				if(has_sys_attribute)
				{
					PROPERTY	new_property	= (PROPERTY)p;
					Add(new_property);
				}
			}

			return true;
		}


		/// <summary>
		/// indexer: get the attribute by its name.
		/// </summary>
		/// <param name="attribute_name"></param>
		/// <returns></returns>
		public PROPERTY this[string property_name]
		{
			get
			{
				if( Count <= 0 || string.IsNullOrEmpty( property_name))
					return null;

				foreach (PROPERTY item in this)
				{
					if( string.Compare( item.Name, property_name, true) == 0)
						return item;
				}
				return null;
			}
		}

		/// <summary>
		/// Helper: remove the attribute by its name.
		/// </summary>
		/// <param name="item">The attribute (or an attribute object having the name of the attribute to be removed)</param>
		/// <returns></returns>
		public new bool Remove(PROPERTY item)
		{
			if( Count <= 0 || item == null)
				return false;

			PROPERTY	my_item	= this[item.Name];

			if( my_item != null)
				return base.Remove(my_item);
			return false;
		}
	}
}
